# This file was automatically generated by Dist::Zilla::Plugin::Test::Pod::Coverage::TrustMe v1.0.1
use strict;
use warnings;
use Test::More;
use Test::Pod::Coverage::TrustMe;

my $config = {};
my $modules = [
  "JSON::Schema::Modern",
  "JSON::Schema::Modern::Annotation",
  "JSON::Schema::Modern::Document",
  "JSON::Schema::Modern::Error",
  "JSON::Schema::Modern::Result",
  "JSON::Schema::Modern::ResultNode",
  "JSON::Schema::Modern::Utilities",
  "JSON::Schema::Modern::Vocabulary",
  "JSON::Schema::Modern::Vocabulary::Applicator",
  "JSON::Schema::Modern::Vocabulary::Content",
  "JSON::Schema::Modern::Vocabulary::Core",
  "JSON::Schema::Modern::Vocabulary::FormatAnnotation",
  "JSON::Schema::Modern::Vocabulary::FormatAssertion",
  "JSON::Schema::Modern::Vocabulary::MetaData",
  "JSON::Schema::Modern::Vocabulary::Unevaluated",
  "JSON::Schema::Modern::Vocabulary::Validation",
];

plan tests => scalar @$modules;

for my $module (@$modules) {
    pod_coverage_ok($module, $config);
}

done_testing;
